#define LETTER(x) ((x >= 'A' && x <= 'Z')?(x+('a'-'A')):(x))
#define UPPERCASE(x) (x >= 'A' && x <= 'Z')
#define UCASE(x) (x &  0xCF)
#define LCASE(x) (x |  0x20)


void CopyString(char *s1, char *s2) {
   int index=-1;
   while (s2[index] != '\0') {
      index++;
      s1[index] = s2[index];
   }
   return;
}

void GetInput(char * &male, char * &female, int &numgenes) {
   int conx=0, cony=0;
   bool fadedin = false, proceed=true;
   RGB_color * pal;

   Set_Video_Mode(VGA256);
   
   Cls();

   Restore_Default_Palette();
   
   pal = Init_BG_Pal();

   Blank_Palette();

   Draw_BG();

   Blit_Con_String("Autosomal Punnett Square Generator", conx, cony, 15, 0, true);
   CR(conx,cony);

   do {
     Blit_Con_String("Enter # of genes. ", conx, cony, 15, 0, true);
     if (!fadedin) {Fade_Into(pal); delete pal; fadedin=true;}
     numgenes = atoi(Input_Con_String(conx, cony, 15, 0, 2,true));
     LINEFEED
     if (numgenes > 3) {
        Blit_Con_String("The punnett square will not fit on the",
                        conx,cony,15,0,true);
        LINEFEED
        Blit_Con_String("screen. Continue anyway (Y/N)? ",conx, cony, 15,0,true);
        while (!kbhit()) {Rotate_Palette();}
        if ((getch() | 0x20) == 'n') {
           proceed = false;
        } else {
           LINEFEED
           Blit_Con_String("You've been warned.",conx,cony,15,0,true);
        }
        LINEFEED
     } else proceed = true;
     
   } while (numgenes < 1 || !proceed);
   
   do {
      Blit_Con_String("Enter male parent genotype. ", conx, cony, 15, 0, true);
      male = Input_Con_String(conx,cony,15,0,numgenes*2,true);
      LINEFEED
   } while (strlen(male) != numgenes*2);

   do {
      Blit_Con_String("Enter female parent genotype. ", conx, cony, 15, 0, true);
      female = Input_Con_String(conx,cony,15,0,numgenes*2,true);
      LINEFEED
   } while (strlen(female) != numgenes*2);

   Fade_Out();

   return;
}

void AlphaOrder(char * str) {
   int x, len=strlen(str)-1;
   bool inorder=false;
   char temp;
   while (!inorder) {
      inorder = true;
      for (x=0; x < len; x++) {
         if (len != 0 && (LETTER(str[x]) > LETTER(str[x+1]))) {
            inorder=false;
            temp = str[x+1];
            str[x+1] = str[x];
            str[x] = temp;
         }
      }
   }
   return;
}

void OrderAlleles(char * str) {
   int x, len=strlen(str)-1;
   bool inorder=false;
   char temp;
   while (!inorder) {
      inorder = true;
      for (x=0; x < len; x++) {
         if (len != 0 && (LETTER(str[x]) > LETTER(str[x+1]))) {
            inorder=false;
            temp = str[x+1];
            str[x+1] = str[x];
            str[x] = temp;
         }
      }
   }

   for (x=0; x < len; x+=2) {
      if (str[x] > str[x+1]) {
         temp = str[x+1];
         str[x+1] = str[x];
         str[x] = temp;
      }
   }
   
   return;
}

void MakeGamete(int numgenes, int numcombos, char ** gamete, char * parent) {
   int x;
   unsigned char cnt, mask;

   char * gametetemp = new char [numgenes+1];

   for (cnt=0; cnt < numcombos; cnt++) {
      mask = 0x01;
      for (x=0; x < numgenes; x++) {
         if (cnt & mask) {
            gametetemp[x] = parent[x*2+1];
         } else {
            gametetemp[x] = parent[x*2];
         }
         mask = mask << 1;
      }
      gametetemp[x] = '\0';
      AlphaOrder(gametetemp);
      gamete[cnt] = new char [numgenes+1];
      CopyString(gamete[cnt],gametetemp);
   }
   return;
}

char * Append(char * first, char * second) {
   int c,x=strlen(first), y=strlen(second);
   char * result = new char [x+y+1];

   CopyString(result, first);

   for (c=0; c < strlen(second); c++) {
      result[c+x] = second[c];
   }

   result[c+x] = '\0';
   return result;
}

//Can only be used for mono and di-hybrids
void Graph_Punnett(int numgenes, int numcombos, int padding, int topspace,
                  char ** sperm, char ** eggs) {
   int hsize, xsize,ysize,x,y,ex,ey;
   int scale;
   char * temp;

   if (numgenes == 1) scale = 4;
   if (numgenes == 2) scale = 2;
   
   hsize = CHAR_WIDTH * scale * numgenes + padding * 2;
   xsize = CHAR_WIDTH * scale * numgenes * 2 + padding * 2;
   ysize = CHAR_HEIGHT * scale + padding * 2;
                                  
   for (y=0; y < numcombos; y++) {
      for (x=0; x < numcombos; x++) {
         ex = x*xsize+hsize;
         ey = y*ysize+ysize;

         Box(ex,ey+topspace,ex+xsize,ey+ysize+topspace,15);
         temp = Append(eggs[x],sperm[y]);
         if (!temp) MemError();
         OrderAlleles(temp);

         Blit_Scaled_String(temp, ex+padding, ey+padding+topspace, 15, 0,
                            true,scale,scale);

         delete temp;
      }
   }

   ex = 0;
   for (y=0; y < numcombos; y++) {
      ey = ysize + y*ysize;
      Blit_Scaled_String(sperm[y], ex+padding, ey+padding+topspace,15,0,true,
                         scale,scale);
   }

   ey = 0;
   for (x=0; x < numcombos; x++) {
      ex = hsize + hsize/2 + x*xsize;
      Blit_Scaled_String(eggs[x], ex+padding, ey+padding+topspace,15,0,true,
                         scale,scale);
   }

   return;
}

void Text_Punnett(int numgenes, int numcombos, char ** sperm, char ** eggs) {
   int x,y,z,size,hsize;
   char * temp;

   hsize = numgenes+1;
   
   cout << "\n ";
   
   //Write list of eggs

   for (x=0; x < hsize; x++) {
      cout << ' ';
   }

   for (x=0; x < numcombos; x++) {
      cout << eggs[x] << ' ';
      for (y=0; y < numgenes; y++) {
         cout << ' ';
      }
   }
   cout << '\n';

   //Write table header

   for (x=0; x < hsize; x++) {
      cout << ' ';
   }

   cout << '';
   
   for (x=0; x < numcombos; x++) {
      for (y=0; y < numgenes; y++) {
         cout << "";
      }
      cout << '';
   }

   cout << "\b\n";

   //Write table body

   for (y=0; y < numcombos; y++) {
      cout << sperm[y] << " ";
      for (x=0; x < numcombos; x++) {
         temp = Append(eggs[x],sperm[y]);
         if (!temp) MemError();
         OrderAlleles(temp);
         cout << temp << "";
         delete temp;
      }
      cout << "\n";
      
      //Every other line is a sperator bar

      if (y != numcombos-1) {

         for (x=0; x < numgenes+1; x++) {
            cout << ' ';
         }
         cout << '';
         for (x=0; x < numcombos; x++) {
            for (z=0; z < numgenes; z++) {
               cout << "";
            }
            cout << '';
         }
         cout << "\b\n";
      }
   }
   
   //Write table footer

   for (x=0; x < hsize; x++) {
      cout << ' ';
   }
   
   cout << '';
   
   for (x=0; x < numcombos; x++) {
      for (y=0; y < numgenes; y++) {
         cout << "";
      }
      cout << '';
   }

   cout << "\b\n";
   
   return;
}

int GCF(int * r, int n) {
   int least=0, c, d, gcf;
   bool willtolive=true;
   for (c=0; c < n; c++) {
      if ((least == 0||r[c] < least) && r[c] !=0) least=r[c];
   }
   
   for (c=least; (c > 0) && willtolive; c--) {
      for (d=0; (d < n) && willtolive; d++) {
         if (r[d] % c != 0) willtolive=false;
      }
      if (willtolive) {
         gcf = c;
         willtolive=false;
      }
   }

   return gcf;
}
   

void Frequencies(char ** eggs, char ** sperm, int numgenes,int numcombos,bool graph) {
   int conx=2,cony=150,x,y,c;
   int * ratio = new int [0x01 << numgenes];
   unsigned char mask, currtot;
   char * temp, * temp2;
   char ** symbols = new char * [0x01 << numgenes];

   for (x=0; x < (0x01 << numgenes); x++) {
      ratio[x] = 0;
      symbols[x] = 0;
   }

   for (y=0; y < numcombos; y++) {
      for (x=0; x < numcombos; x++) {
         temp = Append(eggs[x],sperm[y]);
         temp2 = new char [numgenes+1];
         if (!temp || !temp2) MemError();
         OrderAlleles(temp);
         
         mask = 0x01;
         currtot = 0x00;
         for (c=0; c < numgenes*2-1; c+=2) {
            if (UPPERCASE(temp[c]) || UPPERCASE(temp[c+1])) {
               currtot = currtot + mask;
               temp2[c/2] = UCASE(temp[c]);
            } else {
               temp2[c/2] = LCASE(temp[c]);
            }
            mask = mask << 1;
         }
         temp2[c/2] = 0;
         ratio[currtot]++;         
         if (!symbols[currtot]) { 
            symbols[currtot] = temp2;
         } else {delete temp2;}

         delete temp;
      }
   }

   //Sort
   bool inorder=false;
   
   while (!inorder) {
      inorder=true;
      for (x=0; x < (0x01 << numgenes)-1 ; x++) {
         if (ratio[x] > ratio[x+1]) {
            inorder=false;
            c    = ratio[x];
            temp = symbols[x];
            ratio[x]  = ratio[x+1];
            symbols[x] = symbols[x+1];
            ratio[x+1]  = c;
            symbols[x+1] = temp;
         }
      }
   }

   //Reduce
   y = GCF(ratio, (0x01 << numgenes));
   for (x=0; x < (0x01 << numgenes); x++) {
      ratio[x] /= y;
   }
   
   if (graph) {
      Blit_Con_String("Phenotypic ratio:",conx,cony,15,0,true);
      LINEFEED
      conx=2;
      cony++;
   } else {
      cout << "Phenotypic ratio:\n";
   }
   
   for (x=(0x01 << numgenes)-1; x >= 0; x--) {
      if (ratio[x]) {
         temp = Append(Append(Append(symbols[x],":"), itoa(ratio[x],"0123456789",10)), " ");
         if (graph) Blit_Con_String(temp, conx, cony, 15, 0, true);
         else cout << temp;
      }
   }
   
   delete ratio;

   return;
}
   
   
void Autosomal(void) {
   int numgenes, numcombos, x, y;
   char * male, * female, * temp, males[3]={11,32,0}, females[3]={12,32,0};
   RGB_color * pal;
   
   Set_Video_Mode(VGA256);

   pal = Save_Palette();

   GetInput(male, female, numgenes);

   numcombos = 1 << numgenes;
   
   char ** sperm = new char * [numcombos];
   char ** eggs  = new char * [numcombos];

   if (!sperm || !eggs) MemError();

   MakeGamete(numgenes, numcombos, sperm, male);
   MakeGamete(numgenes, numcombos, eggs, female);

   if (numgenes < 3) {
      Cls();

      Restore_Palette(pal);

      delete pal;

      pal = Init_BG_Pal();

      Blank_Palette();

      Draw_BG();

      temp = Append(Append(Append(Append(males, male)," x "), females), female);

      if (numgenes == 1) {
         Blit_String(temp, 8*CHAR_WIDTH, 0, 15, 0, true);
      } else {
         Blit_String(temp, 14*CHAR_WIDTH, 0, 15, 0, true);
      }

      Graph_Punnett(numgenes, numcombos, 2, 10, sperm, eggs);

      Frequencies(eggs, sperm, numgenes, numcombos, true);
      
      Fade_Into(pal);

      delete pal;

      while (!kbhit()) {Rotate_Palette();}
      getch();

   } else {
      Set_Video_Mode(TEXT_MODE);

      Blank_Palette();
      
      clrscr();

      cout << males << male << " x " << females << female << '\n';

      Text_Punnett(numgenes, numcombos, sperm, eggs);

      Frequencies(eggs, sperm, numgenes, numcombos, false);

      Fade_In_Text();

      PAUSE;
   }

   Fade_Out();

   return;
}
